/// Klotski Engine by Dnomd343 @2024

#pragma once

#include "internal/constant/group.h"
#include "internal/constant/group_union.h"

namespace klotski::group {

class Group;
class CaseInfo;
class GroupCases;
class GroupUnion;

} // namespace klotski::group
